from fonts import FontPrinter, FontRepository
from PIL import Image, ImageDraw
import math

class SeriesEntry:
    def __init__(self, label, values):
        self.label = label
        self.values = values

class ValueConfig:
    def __init__(self):    
        self.axis = 0
        self.color = 'b'
        self.type = 'line'

class LabelConfig:
    def __init__(self):
        self.height = 20
        self.label_margin = 4
        self.label_distance = 4
        self.font = FontRepository.SairaCondensed8
        self.color = "b"

class AxisConfig:
    def __init__(self):
        self.use_min = None
        self.use_max = None
        self.font = FontRepository.SairaCondensed8
        self.width = 20
        self.format = "%.1f"
        self.step = 1
        self.color = "b"
        self.label_distance = 4
        self.label_margin = 4

class Series:
    def __init__(self, entries, value_configs, label_config, axis_config):

        # Check, if parameters are valid

        if len(entries) < 1:
            raise Exception("Invalid count of entries!")
        seriesCount = len(entries[0].values)
        for i in range(1, len(entries)):
            if len(entries[i].values) != seriesCount:
                raise Exception("Every entry must have the same number of values!")
        if len(value_configs) != seriesCount:
            raise Exception("You have to supply same count of ValueConfigs as number of values per entry!")
        
        axis_1_used = any([c.axis == 0 for c in value_configs])
        axis_2_used = any([c.axis == 1 for c in value_configs])

        if axis_1_used and len(axis_config) < 1 or axis_2_used and len(axis_config) < 2:
            raise Exception("Missing axis configuration!")

        # Set members

        self.entries = entries
        self.value_configs = value_configs
        self.label_config = label_config
        self.axis_config = axis_config

class Chart:
    MIN_CHART_WIDTH = 10
    MIN_CHART_HEIGHT = 10

    class Rect:
        def __init__(self, x, y, w, h):
            self.x = x
            self.y = y
            self.w = w
            self.h = h
            self.l = x
            self.t = y
            self.r = x + w
            self.b = y + h

    @staticmethod
    def eval_axis_extreme(series, axis):
        axis_min = series.axis_config[axis].use_min
        axis_max = series.axis_config[axis].use_max
        for i in [x for x in range(0, len(series.value_configs)) if series.value_configs[x].axis == axis]:
            evaluated_min = min(e.values[i] for e in series.entries)
            evaluated_max = max(e.values[i] for e in series.entries)

            axis_min = min(axis_min, evaluated_min) if axis_min != None else evaluated_min
            axis_max = max(axis_max, evaluated_max) if axis_max != None else evaluated_max

        if axis_min == None and axis_max == None:
            return (axis_min, axis_max)

        dist = axis_max - axis_min
        if dist == 0:
            dist = 4

        # Reserve some margin space on top and bottom 
        # But only if no explicit borders are defined

        if series.axis_config[axis].use_min == None:
            axis_min = axis_min - dist / 4
        if series.axis_config[axis].use_max == None:
            axis_max = axis_max + dist / 4

        # Round to step

        axis_min = math.floor(axis_min / series.axis_config[axis].step) * series.axis_config[axis].step
        axis_max = math.ceil(axis_max / series.axis_config[axis].step) * series.axis_config[axis].step

        return (axis_min, axis_max)
        
    @staticmethod
    def map_pixels(low_value, low_pixels, high_value, high_pixels, value):
        return low_pixels + (value - low_value) * ((high_pixels - low_pixels) / (high_value - low_value))

    @staticmethod
    def draw_axis(axis_min, axis_max, axis_rect, axis_config, black_image, red_image, alignleft):
        
        black_draw = ImageDraw.Draw(black_image)
        red_draw = ImageDraw.Draw(red_image)

        if alignleft:
            if axis_config.color == 'b':
                black_draw.line([(axis_rect.l, axis_rect.t), (axis_rect.l, axis_rect.b-1)], fill = 0, width = 1)
            else:
                red_draw.line([(axis_rect.l, axis_rect.t), (axis_rect.l, axis_rect.b-1)], fill = 0, width = 1)
        else:
            if axis_config.color == 'b':
                black_draw.line([(axis_rect.r - 1, axis_rect.t), (axis_rect.r - 1, axis_rect.b-1)], fill = 0, width = 1)
            else:
                red_draw.line([(axis_rect.r - 1, axis_rect.t), (axis_rect.r - 1, axis_rect.b-1)], fill = 0, width = 1)
        
        value = axis_min
            
        last_y = None

        while value <= axis_max:
            value_text = axis_config.format % value
            (twidth, theight) = FontPrinter.measure(value_text, axis_config.font)

            new_y = int(Chart.map_pixels(axis_min, axis_rect.b - 1, axis_max, axis_rect.t, value) - theight / 2)
            if new_y < axis_rect.t:
                break

            if last_y == None or new_y + theight + axis_config.label_distance < last_y:
                # Drawing new entry

                if alignleft:
                    FontPrinter.print_text(black_image if axis_config.color == "b" else red_image, (axis_rect.l + axis_config.label_margin, new_y), value_text, axis_config.font)
                else:
                    FontPrinter.print_text(black_image if axis_config.color == "b" else red_image, (axis_rect.r - axis_config.label_margin - twidth, new_y), value_text, axis_config.font)
                last_y = new_y

            value = value + axis_config.step

    @staticmethod
    def draw(series, black_image, red_image, position, size):
        # Checks

        (laxis_min, laxis_max) = Chart.eval_axis_extreme(series, 0)
        use_laxis = laxis_min != None and laxis_max != None
        (raxis_min, raxis_max) = Chart.eval_axis_extreme(series, 1)
        use_raxis = raxis_min != None and raxis_max != None

        (chart_width, chart_height) = size
        required_width = (series.axis_config[0].width if use_laxis else 0) + Chart.MIN_CHART_WIDTH + (series.axis_config[1].width if use_raxis else 0)
        if chart_width < required_width:
            raise Exception("Size is not wide enough!")

        required_height = series.label_config.height + Chart.MIN_CHART_HEIGHT
        if chart_height < required_height:
            raise Exception("Size is not high enought!")

        # Prepare buffers
        
        black_buffer = Image.new('1', size, 255)
        black_draw = ImageDraw.Draw(black_buffer)
        red_buffer = Image.new('1', size, 255)
        red_draw = ImageDraw.Draw(red_buffer)

        # Evaluate regions

        vaxis_height = chart_height - series.label_config.height
        laxis_rect = Chart.Rect(0, 0, series.axis_config[0].width, vaxis_height) if use_laxis else Chart.Rect(0, 0, 0, vaxis_height)
        raxis_rect = Chart.Rect(chart_width - series.axis_config[1].width, 0, series.axis_config[1].width, vaxis_height) if use_raxis else Chart.Rect(chart_width, 0, 0, vaxis_height)
        label_axis_rect = Chart.Rect(laxis_rect.r, vaxis_height, raxis_rect.l - laxis_rect.r + 1, series.label_config.height)
        chart_rect = Chart.Rect(laxis_rect.r, 0, raxis_rect.l - laxis_rect.r, vaxis_height)

        # Draw left axis + left line
        if use_laxis:
            Chart.draw_axis(laxis_min, laxis_max, laxis_rect, series.axis_config[0], black_buffer, red_buffer, False)
            
        # Draw right axis + right line
        if use_raxis:
            Chart.draw_axis(raxis_min, raxis_max, raxis_rect, series.axis_config[1], black_buffer, red_buffer, True)

        # Draw labels + bottom line

        if series.label_config.color == 'b':
            black_draw.line([(label_axis_rect.l, label_axis_rect.t), (label_axis_rect.r - 1, label_axis_rect.t)], fill = 0, width = 1)
        else:
            red_draw.line([(label_axis_rect.l, label_axis_rect.t), (label_axis_rect.r - 1, label_axis_rect.t)], fill = 0, width = 1)

        last_x = None
        last_w = None
        for i in range(len(series.entries)):
            entry = series.entries[i]
            value_text = str(entry.label)

            (text_w, text_h) = FontPrinter.measure(value_text, series.label_config.font)

            new_x = int(Chart.map_pixels(0, label_axis_rect.l, len(series.entries) - 1, label_axis_rect.r - 1, i) - text_w / 2)

            if last_x == None or last_x + last_w + series.label_config.label_distance <= new_x:
                # Drawing new entry

                FontPrinter.print_text(black_buffer if series.label_config.color == 'b' else red_buffer, (new_x, label_axis_rect.t + series.label_config.label_margin), value_text, series.label_config.font)

                last_x = new_x
                last_w = text_w

        # Draw series

        for value_index in range(len(series.value_configs)):
            value_config = series.value_configs[value_index]        
            (min_value, max_value) = (laxis_min, laxis_max) if value_config.axis == 0 else (raxis_min, raxis_max)

            if value_config.type == "line":
                draw = black_draw if value_config.color == "b" else red_draw
                line = []
                for i in range(len(series.entries)):
                    x = Chart.map_pixels(0, chart_rect.l, len(series.entries) - 1, chart_rect.r - 1, i)
                    y = Chart.map_pixels(min_value, chart_rect.b - 1, max_value, chart_rect.t, series.entries[i].values[value_index])
                    line.append((x, y))
                draw.line(line, fill = 0, width = 1, joint="curve")

        # Render chart on given draws

        black_image.paste(black_buffer, position)
        red_image.paste(red_buffer, position)